/*
 * Copyright © 2020, Luis Andrés Lange <https://javacomm.net>
 *
 * This Source Code Form is subject to the terms of the Mozilla Public
 * License, v. 2.0. If a copy of the MPL was not distributed with this
 * file, You can obtain one at http://mozilla.org/MPL/2.0/.
 */
package net.javacomm.client.info;

import java.awt.Dimension;
import java.awt.FlowLayout;
import java.beans.PropertyChangeListener;
import java.beans.PropertyChangeSupport;
import java.beans.PropertyVetoException;
import java.io.BufferedReader;
import java.io.IOException;
import java.io.InputStream;
import java.io.InputStreamReader;
import java.net.URL;
import javax.swing.Box;
import javax.swing.BoxLayout;
import javax.swing.ImageIcon;
import javax.swing.JEditorPane;
import javax.swing.JLabel;
import javax.swing.JPanel;
import javax.swing.JScrollPane;
import javax.swing.JTextArea;
import javax.swing.event.HyperlinkEvent;
import javax.swing.event.HyperlinkListener;
import javax.swing.event.InternalFrameAdapter;
import javax.swing.event.InternalFrameEvent;
import javax.swing.text.html.HTMLDocument;
import javax.swing.text.html.HTMLFrameHyperlinkEvent;
import org.apache.logging.log4j.LogManager;
import org.apache.logging.log4j.Logger;
import net.javacomm.client.resource.Resource;
import net.javacomm.multilingual.MultilingualButton;
import net.javacomm.multilingual.MultilingualLabel;
import net.javacomm.multilingual.schema.ISO639;
import net.javacomm.multilingual.schema.KEY;
import net.javacomm.share.Constants;
import net.javacomm.window.manager.Control;
import net.javacomm.window.manager.Frames;
import net.javacomm.window.manager.WM;
import net.javacomm.window.manager.WindowManagerInternalFrame;



/*****************************************************************************/
/*                                                                           */
/*                          Class JInfoPane                                  */
/*                                                                           */
/*****************************************************************************/

public class JInfoPane extends WindowManagerInternalFrame implements HyperlinkListener {
  private static final long serialVersionUID = 8033663834096696319L;
  private final static Logger log = LogManager.getLogger(JInfoPane.class);
  public final static Dimension PANEL_SIZE = new Dimension(420, 540);
  private JPanel mainPane = new JPanel();
  private JPanel panSouth = new JPanel(new FlowLayout(FlowLayout.CENTER));
  private JPanel panLogo = new JPanel(new FlowLayout(FlowLayout.CENTER));
  private JPanel panVersion = new JPanel(new FlowLayout(FlowLayout.CENTER));
  private Class<? extends JInfoPane> resource;
  private MultilingualLabel labelVersion = new MultilingualLabel(KEY.LABEL_VERSION);
  private JLabel labelNexus = new JLabel();
  private MultilingualButton buttonSchliessen = new MultilingualButton(KEY.BUTTON_SCHLIESSEN);
  private InternalAction internalAction = new InternalAction();
  private PropertyChangeSupport changes = new PropertyChangeSupport(this);
  private JScrollPane scroll = new JScrollPane();
  private JTextArea editor = new JTextArea();
  private BoxLayout box = new BoxLayout(mainPane, BoxLayout.Y_AXIS);
  private InputStream inStream;

  public JInfoPane() {
    mainPane.setLayout(box);
    addInternalFrameListener(internalAction);
    setContentPane(mainPane);
    resource = getClass();

    mainPane.add(panVersion);
    panVersion.add(labelVersion);

    mainPane.add(Box.createRigidArea(new Dimension(0, 12)));
    mainPane.add(panLogo);
    URL url = resource.getResource(Resource.MPL2);

    labelNexus.setIcon(new ImageIcon(url));

    panLogo.add(labelNexus);

    scroll.setViewportView(editor);
    scroll.setBorder(null);

    BufferedReader bufferedReader = null;
    try {
      inStream = resource.getResourceAsStream(Resource.LICENSE);

      InputStreamReader instreamreader = new InputStreamReader(inStream, "UTF-8");
      bufferedReader = new BufferedReader(instreamreader);
      editor.setText(new String());
      while (bufferedReader.ready()) {
        String read = bufferedReader.readLine();
        editor.append(read);
        editor.append("\r\n");
      }
      editor.setCaretPosition(0);
      editor.setEditable(false);
      editor.setLineWrap(true);
      editor.setWrapStyleWord(true);

    }
    catch (IOException e) {
      log.info(e.getMessage(), e.getCause());
    }
    finally {
      try {
        if (bufferedReader != null) bufferedReader.close();
      }
      catch (IOException e) {}
    }
    // mainPane.add (Box.createRigidArea(new Dimension (0,12)));
    mainPane.add(scroll);

    buttonSchliessen.addActionListener(event -> {
      doDefaultCloseAction();
    });
    buttonSchliessen.setFocusPainted(false);
    mainPane.add(Box.createVerticalGlue());
    mainPane.add(panSouth);
    panSouth.add(buttonSchliessen);

  }



  @Override
  public void activated() {
    try {
      setSelected(true);
    }
    catch (PropertyVetoException e) {}
  }



  public void addInfoListener(PropertyChangeListener l) {
    changes.addPropertyChangeListener(l);
  }



  @Override
  public Dimension getMinimumSize() {
    return new Dimension(350, 450);
  }



  @Override
  public Frames getType() {
    return Frames.INFO;
  }



  @Override
  public void hyperlinkUpdate(HyperlinkEvent event) {
    HTMLDocument doc;
    JEditorPane pane;

    if (event.getEventType().equals(HyperlinkEvent.EventType.ACTIVATED)) {
      pane = (JEditorPane) event.getSource();
      if (event instanceof HTMLFrameHyperlinkEvent evt) {
        doc = (HTMLDocument) pane.getDocument();
        doc.processHTMLFrameHyperlinkEvent(evt);
      }
      else {
        try {
          pane.setPage(event.getURL());
        }
        catch (IOException e) {
          log.info(e.getMessage(), e.getCause());
        }
      }
    }
  }



  @Override
  public void insertWM(WM frame) {
    // nichts machen
  }



  public void removeInfoListener(PropertyChangeListener l) {
    changes.removePropertyChangeListener(l);
  }



  @Override
  public void removeWM(WM frame) {
    // nichts machen
  }



  @Override
  public void setLanguage(ISO639 code) {
    buttonSchliessen.setLanguage(code);
    setLanguage(KEY.FRAME_INFO, code);
    labelVersion.setLanguage(code);
    String neuerText = labelVersion.getText().replace("XXX", Constants.VERSION);
    labelVersion.setText(neuerText);
  }

  /*****************************************************************************/
  /*                                                                           */
  /* Class InternalAction */
  /*                                                                           */
  /*****************************************************************************/
  class InternalAction extends InternalFrameAdapter {

    @Override
    public void internalFrameActivated(InternalFrameEvent event) {
      mainPane.revalidate();
      buttonSchliessen.requestFocus();
    }



    @Override
    public void internalFrameClosed(InternalFrameEvent event) {
      changes.firePropertyChange("JInfoPane", Control.NULL, Control.QUIT);
    }

  }

}
