/**
 * Copyright © 2022, Luis Andrés Lange <https://javacomm.net>
 *
 * This Source Code Form is subject to the terms of the Mozilla Public
 * License, v. 2.0. If a copy of the MPL was not distributed with this
 * file, You can obtain one at http://mozilla.org/MPL/2.0/.
 */
package net.javacomm.client.screencaster;

import jakarta.ws.rs.ProcessingException;
import jakarta.ws.rs.client.Client;
import jakarta.ws.rs.client.ClientBuilder;
import jakarta.ws.rs.client.Entity;
import jakarta.ws.rs.core.MediaType;
import jakarta.ws.rs.core.Response;
import java.awt.AWTException;
import java.awt.EventQueue;
import java.awt.Graphics2D;
import java.awt.MouseInfo;
import java.awt.Point;
import java.awt.Rectangle;
import java.awt.Robot;
import java.awt.Toolkit;
import java.awt.image.BufferedImage;
import java.io.ByteArrayOutputStream;
import java.io.IOException;
import java.lang.reflect.InvocationTargetException;
import java.util.concurrent.ArrayBlockingQueue;
import java.util.concurrent.ExecutionException;
import java.util.concurrent.Future;
import java.util.concurrent.TimeUnit;
import java.util.concurrent.TimeoutException;
import javax.imageio.ImageIO;
import org.apache.logging.log4j.LogManager;
import org.apache.logging.log4j.Logger;
import org.glassfish.jersey.media.multipart.FormDataMultiPart;
import org.glassfish.jersey.media.multipart.MultiPartFeature;
import net.javacomm.client.resource.Resource;
import net.javacomm.share.Constants;



public class ImageCaster {

  private final static Logger log = LogManager.getLogger(ImageCaster.class);
  private final static int COUNT_THREADS = 2;
  private Robot robot;
  private Rectangle screenSize;
  private ArrayBlockingQueue<byte[]> queue = new ArrayBlockingQueue<>(COUNT_THREADS);
  private Producer p1;
  private Consumer c1;

  private Thread producer_1;
  private Thread consumer_1;

  private static ImageCaster caster;

  private ImageCaster() throws AWTException {
    robot = new Robot();
    screenSize = new Rectangle(Toolkit.getDefaultToolkit().getScreenSize());
  }


  static {
    try {
      caster = new ImageCaster();
    }
    catch (AWTException e) {
      log.fatal(e.getMessage(), e);
    }
  }


  public static synchronized ImageCaster getInstance() {
    return caster;
  }




  /**
   * Die Aufzeichnung von Screenshots beginnt.
   * 
   * 
   * @param senderUserid
   *                     die Screenshots stammen von diesem User
   */
  public void recordScreenshots(String senderUserid) {
    stopRecording();

    p1 = new Producer(queue, robot, screenSize);
    c1 = new Consumer(queue, senderUserid); // userid ist der Sender

    producer_1 = new Thread(Thread.currentThread().getThreadGroup(), p1, "screenshot");
    consumer_1 = new Thread(Thread.currentThread().getThreadGroup(), c1, "sendToDatabase");

    consumer_1.start();
    producer_1.start();


  }



  /**
   * 
   * Screenshots werden nicht mehr aufgezeichnet und versendet. Ein mehrmaliger
   * Aufruf bewirkt nichts Neues.
   * 
   */
  public void stopRecording() {
    if (consumer_1 != null) consumer_1.interrupt();
    if (producer_1 != null) producer_1.interrupt();
    try {
      if (consumer_1 != null) consumer_1.join();
    }
    catch (InterruptedException e) {
      log.error(e.getMessage(), e);
    }
    try {
      if (producer_1 != null) producer_1.join();
    }
    catch (InterruptedException e) {
      log.error(e.getMessage(), e);
    }
  }




  /**
   * Über den Consumer werden die Screenshots an den Server gesendet.
   * 
   * 
   * @author llange
   *
   */
  private static class Consumer implements Runnable {

    private ArrayBlockingQueue<byte[]> queue;
    private String userid;

    /**
     * Der Consumer wird initialisiert.
     * 
     * @param queue
     *               aus dieser Queue werden die Bilder gelesen.
     * @param userid
     *               an diesen Anwender werden die Bilder gesendet
     */
    public Consumer(ArrayBlockingQueue<byte[]> queue, String userid) {
      this.queue = queue;
      this.userid = userid;
    }



    @Override
    public void run() {
      Client client = null;
      while (!Thread.currentThread().isInterrupted()) {
        client = ClientBuilder.newClient();
        client.register(MultiPartFeature.class);
        try(FormDataMultiPart formMultipart = new FormDataMultiPart()) {

          byte[] picture = queue.take(); // aus der Queue lesen

          formMultipart.field("userid", userid);
          formMultipart.field("image", picture, MediaType.APPLICATION_OCTET_STREAM_TYPE);

          StringBuilder url = new StringBuilder(Constants.HTTP).append(Constants.DOMAIN)
              .append("/javacommserver/screenshot/save");

          /**
           * Das Bild wird asynchron an den Server gesendet.
           * 
           */
          Future<Response> response = client.target(url.toString()).request(MediaType.APPLICATION_JSON)
              .async()
              .post(Entity.entity(formMultipart, formMultipart.getMediaType()));

          response.get(200, TimeUnit.MILLISECONDS); // 5 Bilder pro Sekunde


        }
        catch (ProcessingException | IOException e) {
          // Connection kann verloren gehen
          log.warn(e.getMessage());
        }
        catch (InterruptedException e) {
          break;
        }
        catch (ExecutionException e) {
          log.warn(e.getMessage(), e.getCause());
        }
        catch (TimeoutException e) {
//          log.error(e.getMessage(), e.getCause());
        }
        finally {
          client.close();
        }
      }
    }

  }


  /**
   * Der Producer erzeugt Screenshots im JPG-Format.
   * 
   * 
   * @author llange
   *
   */
  private static class Producer implements Runnable {

    private Robot robot;
    private Rectangle screenSize;
    private ArrayBlockingQueue<byte[]> queue;
    private BufferedImage duke = null;
    BufferedImage image;
    ByteArrayOutputStream outputStream;

    /**
     * Der Producer wird initialisiert.
     * 
     * 
     * @param queue
     *                   in diese Queue werden die Screenshots geschrieben
     * @param robot
     * @param screenSize
     *                   die Bildschirmgröße
     */
    public Producer(ArrayBlockingQueue<byte[]> queue, Robot robot, Rectangle screenSize) {
      this.queue = queue;
      this.robot = robot;
      this.screenSize = screenSize;
      outputStream = new ByteArrayOutputStream();
      ImageIO.setUseCache(false);
      try {
        duke = ImageIO.read(getClass().getResource(Resource.DUKEMAUS));
      }
      catch (IOException e) {
        log.warn("Mauscursor konnte nicht gelesen werden");
      }

    }




    @Override
    public void run() {
      if (duke == null) return;
      while (!Thread.currentThread().isInterrupted()) {

        try {
          EventQueue.invokeAndWait(() -> {
            image = robot.createScreenCapture(screenSize);
          });
        }
        catch (InvocationTargetException | InterruptedException e1) {
          return;
        }

        Graphics2D g2 = image.createGraphics();
        Point mousepointer = MouseInfo.getPointerInfo().getLocation();
        g2.drawImage(duke, mousepointer.x, mousepointer.y, null);
        g2.dispose();

        try {
          outputStream.reset();
          ImageIO.write(image, "jpg", outputStream);
          image.flush();
          queue.put(outputStream.toByteArray());
        }
        catch (IOException e) {
          log.error("ImageIO - " + e.getMessage());
        }
        catch (InterruptedException e) {
          return;
        }
      }
    }
  }


}

