/**
 * Copyright © 2022, Luis Andrés Lange <https://javacomm.net>
 *
 * This Source Code Form is subject to the terms of the Mozilla Public
 * License, v. 2.0. If a copy of the MPL was not distributed with this
 * file, You can obtain one at http://mozilla.org/MPL/2.0/.
 */
package net.javacomm.client.screencaster;

import jakarta.ws.rs.NotFoundException;
import jakarta.ws.rs.ProcessingException;
import jakarta.ws.rs.client.Client;
import jakarta.ws.rs.client.ClientBuilder;
import jakarta.ws.rs.core.MediaType;
import java.awt.image.BufferedImage;
import java.io.BufferedInputStream;
import java.io.ByteArrayInputStream;
import java.io.ByteArrayOutputStream;
import java.io.IOException;
import java.io.InputStream;
import java.util.Collections;
import java.util.LinkedList;
import java.util.List;
import java.util.concurrent.ArrayBlockingQueue;
import javax.imageio.ImageIO;
import org.apache.logging.log4j.LogManager;
import org.apache.logging.log4j.Logger;
import net.javacomm.share.Constants;



public class ImageReceiver {

  /**
   * Der Consumer schickt das Bild über einen Listener an andere Objekte. Der
   * Listener läuft in einem eiegenen Thread.
   * 
   * 
   * @author llange
   *
   */
  private class Consumer implements Runnable {
    private ArrayBlockingQueue<byte[]> queue;

    public Consumer(ArrayBlockingQueue<byte[]> queue) {
      this.queue = queue;
    }



    @Override
    public void run() {

      while (!Thread.currentThread().isInterrupted()) {
        try {
          byte[] picture = queue.take();
          if (picture.length == 0) continue;
          BufferedImage image = ImageIO.read(new ByteArrayInputStream(picture));
          ImageReceiverEvent event = new ImageReceiverEvent(this);
          event.setImageReceiverEvent(image);
          for (ImageReceiverListener tmp : listener) {
            tmp.onImage(event);
          }
        }
        catch (InterruptedException | IOException e) {
          return;
        }

      }
    }

  }

  /**
   * Der Producer liest die Screenshots über das Netzwerk aus der Datenbank.
   * 
   * @author llange
   *
   */
  private static class Producer implements Runnable {

    private ArrayBlockingQueue<byte[]> queue;
    private Client client;
    private String userid;

    public Producer(ArrayBlockingQueue<byte[]> queue, String userid) {
      this.userid = userid;
      this.queue = queue;
    }



    @Override
    public void run() {
      ByteArrayOutputStream outbuffer = new ByteArrayOutputStream();
      byte[] buffer = new byte[8192];
      int read;

      while (!Thread.currentThread().isInterrupted()) {
        client = ClientBuilder.newClient();
        StringBuilder url = new StringBuilder(Constants.HTTP).append(Constants.DOMAIN)
            .append("/javacommserver/screenshot/load");

        try(
          InputStream result = client.target(url.toString()).queryParam("userid", userid)
              .request(MediaType.APPLICATION_OCTET_STREAM).get(InputStream.class);
          BufferedInputStream inbuffer = new BufferedInputStream(result);
        ) {

          outbuffer.reset();
          while ((read = inbuffer.read(buffer)) > 0) {
            outbuffer.write(buffer, 0, read);
          }
          queue.put(outbuffer.toByteArray());
        }
        catch (ProcessingException | NotFoundException | IOException e) {
          log.warn(e.getMessage());
        }
        catch (InterruptedException e) {
          break;
        }
        finally {
          client.close();
        }
      }
      log.info("Producer finalized");
    }

  }

  private final static int COUNT_THREADS = 2;
  private final static Logger log = LogManager.getLogger(ImageCaster.class);
  private ArrayBlockingQueue<byte[]> queue = new ArrayBlockingQueue<>(COUNT_THREADS);

  private Producer readFromDatabase_1;
  private Consumer sendToAWT_1;

  private Thread producer1;

  private Thread consumer1;

  private List<ImageReceiverListener> listener = Collections
      .synchronizedList(new LinkedList<ImageReceiverListener>());

  /**
   * Ein Bildschrimempfänger wird erzeugt.
   * 
   */
  public ImageReceiver() {

  }



  /**
   * GUI-Komponenten können sich registrieren, um JPG-Bilder zu empfangen.
   * 
   * 
   * @param receiverListener
   *                         ein Listener
   */
  public synchronized void addImageReceiverListener(ImageReceiverListener receiverListener) {
    listener.add(receiverListener);
  }



  /**
   * Erst nach dem Aufruf von {@link #receive(String)} werden Bilder empfangen.
   * Ein mehrmaliger Aufruf bewirkt nichts.
   * 
   * 
   * @param userid
   *               die Bilder stammen von diesem Anwender
   */
  public void play(String userid) {
    stopPlaying();
    sendToAWT_1 = new Consumer(queue);
    readFromDatabase_1 = new Producer(queue, userid);

    producer1 = new Thread(Thread.currentThread().getThreadGroup(), readFromDatabase_1, "readFromDatabase");
    producer1.start();

    consumer1 = new Thread(Thread.currentThread().getThreadGroup(), sendToAWT_1, "sendToAWT");
    consumer1.start();
  }



  /**
   * Entferne alle Listener auf einen Schlag.
   * 
   */
  public synchronized void removeAllListener() {
    for (ImageReceiverListener tmp : listener) {
      removeImageReceiverListener(tmp);
    }
  }



  /**
   * GUI-Komponenten können sich abmelden, um keine weiteren Bilder zu erhalten.
   * 
   * @param receiverListener
   *                         ein Listener
   */
  public synchronized void removeImageReceiverListener(ImageReceiverListener receiverListener) {
    listener.remove(receiverListener);
  }



  /**
   * Die Bildwiedergabe wird angehalten. Ein mehrmaliger Aufruf hintereinander
   * bewirkt nichts. Alles Resourcen werden freigegeben.
   * 
   */
  public void stopPlaying() {
    if (consumer1 != null) consumer1.interrupt();
    if (producer1 != null) producer1.interrupt();
    try {
      if (consumer1 != null) consumer1.join();
    }
    catch (InterruptedException e) {
      log.error(e.getMessage(), e);
    }
    try {
      if (producer1 != null) producer1.join();
    }
    catch (InterruptedException e) {
      log.error(e.getMessage(), e);
    }
  }

}
