/*
 * Copyright © 2022, Luis Andrés Lange <https://javacomm.net>
 *
 * This Source Code Form is subject to the terms of the Mozilla Public
 * License, v. 2.0. If a copy of the MPL was not distributed with this
 * file, You can obtain one at http://mozilla.org/MPL/2.0/.
 */
package net.javacomm.client.screencaster;

import java.awt.BorderLayout;
import java.awt.Dimension;
import java.awt.EventQueue;
import java.awt.Graphics;
import java.awt.Rectangle;
import java.awt.event.ComponentAdapter;
import java.awt.event.ComponentEvent;
import java.awt.event.MouseAdapter;
import java.awt.event.MouseEvent;
import java.awt.image.BufferedImage;
import java.beans.PropertyChangeEvent;
import java.beans.PropertyChangeListener;
import java.beans.PropertyChangeSupport;
import java.beans.PropertyVetoException;
import java.net.URL;
import javax.swing.ImageIcon;
import javax.swing.JComponent;
import javax.swing.JLabel;
import javax.swing.JScrollBar;
import javax.swing.JScrollPane;
import javax.swing.JSplitPane;
import javax.swing.ScrollPaneConstants;
import javax.swing.UIManager;
import javax.swing.WindowConstants;
import javax.swing.event.InternalFrameAdapter;
import javax.swing.event.InternalFrameEvent;
import org.apache.logging.log4j.LogManager;
import org.apache.logging.log4j.Logger;
import net.javacomm.client.resource.Resource;
import net.javacomm.multilingual.schema.ISO639;
import net.javacomm.multilingual.schema.KEY;
import net.javacomm.window.manager.Control;
import net.javacomm.window.manager.Frames;
import net.javacomm.window.manager.WM;
import net.javacomm.window.manager.WindowAnrufenEvent;
import net.javacomm.window.manager.WindowAnrufenListener;
import net.javacomm.window.manager.WindowManagerInternalFrame;



public class ProjektorInternalFrame extends WindowManagerInternalFrame implements WindowAnrufenListener {

  /**
   * 
   */
  private static final long serialVersionUID = -783434136934302532L;
  private final static Logger log = LogManager.getLogger(ProjektorInternalFrame.class);
  public final static int SCREEN_WIDTH = 871; // 640 + 271
  public final static int SCREEN_HEIGHT = 384; // 360+24
  private ImageIcon image;
  private PropertyChangeSupport changes = new PropertyChangeSupport(this);
  private InternalFrameAdapter internalFrameAdapter = new InternalFrameAdapter() {

    @Override
    public void internalFrameClosing(InternalFrameEvent e) {
      changes.firePropertyChange(ProjektorInternalFrame.class.getName(), Control.NULL, Control.CLOSE);
    }

  };

  private Videosurface videosurface;

  private BorderLayout borderlayout = new BorderLayout();
  private JScrollPane scrollvideopane = new JScrollPane();
  private JLabel corner;
  private JSplitPane splitpaneHorizontal = new JSplitPane(JSplitPane.HORIZONTAL_SPLIT);
  private ComponentAdapter componentAdapter;
  private MouseAdapter mouseAdapter;
  private JScrollBar yScrollbar;
  private JScrollBar xScrollbar;

  public ProjektorInternalFrame() {
    setMaximizable(true);
    setResizable(true);
    setIconifiable(false);
    setClosable(true);
    setDefaultCloseOperation(WindowConstants.DISPOSE_ON_CLOSE);
    URL url = getClass().getResource(Resource.PROJEKTOR_34x26);
    image = new ImageIcon(url);
    setFrameIcon(image);
    setSize(ProjektorInternalFrame.SCREEN_WIDTH, ProjektorInternalFrame.SCREEN_HEIGHT);
    addInternalFrameListener(internalFrameAdapter);
    componentAdapter = new ComponentAdapter() {

      @Override
      public void componentResized(ComponentEvent event) {
        if (ProjektorInternalFrame.this.isMaximum()) {
          Rectangle rectangle = ProjektorInternalFrame.this.getBounds();
          int breite = (int) rectangle.getSize().getWidth() - WindowManagerInternalFrame.SPLITWIDTH
              - WindowManagerInternalFrame.MALUS;
          setHorizontalDivider(breite);
        }
        else {}
      }

    };
    addComponentListener(componentAdapter);
    init();
  }



  private void init() {
    setLayout(borderlayout);
    add(splitpaneHorizontal, BorderLayout.CENTER);
    splitpaneHorizontal.setLeftComponent(scrollvideopane);
    splitpaneHorizontal.setRightComponent(wm);
    splitpaneHorizontal.setDividerLocation(SCREEN_WIDTH - WindowManagerInternalFrame.SPLITWIDTH);
    splitpaneHorizontal.setOneTouchExpandable(true);
    this.setHorizontalDivider(1.0);
    wm.setMinimumSize(new Dimension(0, 0));
    wm.addWMListener(new PropertyChangeListener() {
      @Override
      public void propertyChange(PropertyChangeEvent event) {
        changes.firePropertyChange(event.getPropertyName(), Control.NULL, Control.SELECTED);
      }
    });

    mouseAdapter = new MouseAdapter() {
      @Override
      public void mouseExited(MouseEvent event) {
        if (event.getSource().equals(videosurface)) {
          setHorizontalDivider(
              ProjektorInternalFrame.this.getWidth() - WindowManagerInternalFrame.SPLITWIDTH
          );
        }
      }



      @Override
      public void mouseEntered(MouseEvent event) {
        if (event.getSource().equals(videosurface)) {
          setHorizontalDivider(1.0);
        }
      }

    };

    videosurface = new Videosurface();

    xScrollbar = scrollvideopane.getHorizontalScrollBar();
    xScrollbar.setUnitIncrement(12);

    yScrollbar = scrollvideopane.getVerticalScrollBar();
    yScrollbar.setUnitIncrement(12);

    corner = new JLabel();
    corner.setBackground(UIManager.getColor("nimbusSelectionBackground"));
    corner.setOpaque(true);
    scrollvideopane.setCorner(ScrollPaneConstants.LOWER_RIGHT_CORNER, corner); // ok!
    scrollvideopane.setViewportView(videosurface);
  }



  @Override
  public Dimension getMinimumSize() {
    return new Dimension(ProjektorInternalFrame.SCREEN_WIDTH, ProjektorInternalFrame.SCREEN_HEIGHT);
  }



  /**
   * Die Trennleiste wird in horizontaler Richtung positioniert. Die Angabe
   * erfolgt in Prozent. 0-Prozent ist ganz links und 100-Prozent ist ganz rechts.
   * 
   * @param prozent
   *                liegt im Intervall [0..1]
   */
  public void setHorizontalDivider(double prozent) {
    if (0 > prozent || prozent > 1) {
      log.warn(prozent + " - die Prozentangabe liegt nicht im Intervall [0..1]");
    }
    splitpaneHorizontal.setDividerLocation(prozent);
  }



  /**
   * Die Trennleiste wird in horizontaler Richtung positioniert. Die Angabe
   * erfolgt in Pixel. 0 ist ganz links und {@code SCREEN_WIDTH} ist ganz rechts.
   * 
   * 
   * @param pixel
   *              die X-Position
   */
  public void setHorizontalDivider(int pixel) {
    splitpaneHorizontal.setDividerLocation(pixel);
  }



  /**
   * Alle Resourcen vom VLC Player werden freigegeben.
   * 
   */
  public void shutdown() {
    videosurface.shutdown();
  }



  /**
   * Das Fenster wird geschlossen.
   * 
   */
  public void closeWindow() {
    try {
      setClosed(true);
    }
    catch (PropertyVetoException e) {}
  }



  public void connect(String userid) {
    videosurface.connect(userid);
  }



  public void addProjektorListener(PropertyChangeListener listener) {
    changes.addPropertyChangeListener(listener);
  }



  public void removedProjektorListener(PropertyChangeListener listener) {
    changes.removePropertyChangeListener(listener);
  }



  public void removeAllListener() {
    for (PropertyChangeListener listener : changes.getPropertyChangeListeners()) {
      removedProjektorListener(listener);
    }
    wm.removeAllListener();
    removeComponentListener(componentAdapter);
    videosurface.removeMouseListener(mouseAdapter);
  }



  @Override
  public void activated() {
    try {
      setSelected(true);
      wm.setActivNode(getFrameTitleId());
    }
    catch (PropertyVetoException e) {}
  }



  @Override
  public String getFrameTitleId() {
    return getTitle();
  }



  @Override
  public Frames getType() {
    return Frames.PROJEKTOR;
  }



  @Override
  public void insertWM(WM value) {
    wm.addWindow(value);
  }



  @Override
  public void removeWM(WM frame) {
    wm.removeWM(frame);
  }



  @Override
  public void setFrameTitleId(String id) {
    setTitle(id);
  }



  @Override
  public void setType(Frames value) {
    log.warn("Der Typ wurde mit Frames.PROJEKTOR statisch vorgegeben.");
  }



  @Override
  public void setLanguage(ISO639 code) {
    setLanguage(KEY.FRAME_PROJEKTOR, code); // der neue Name wird gesetzt
    treeView(code);
  }



  @Override
  public void ringRing(WindowAnrufenEvent event) {
    extText = event.getNickname();
    getMultiAbnehmen().setLanguage(KEY.FRAME_ABNEHMEN, event.getCode(), extText);
    getMultiAnrufen().setLanguage(KEY.TELEFON_ANRUFEN, event.getCode(), extText);
  }

  /**
   * 
   * In diesem Panel wird das Video angezeigt.
   * 
   * @author llange
   *
   */
  private static class Videosurface extends JComponent {

    /**
     * 
     */
    private static final long serialVersionUID = 7454789357550113183L;
    ImageReceiver imageReceiver;
    BufferedImage fullscreen;

    public Videosurface() {
      setLayout(new BorderLayout());
      imageReceiver = new ImageReceiver();
    }



    public void connect(String userid) {
      imageReceiver.addImageReceiverListener((event) -> {
        EventQueue.invokeLater(() -> {
          fullscreen = event.getPicture();
          repaint();
        });

      });
      imageReceiver.play(userid);
    }



    public void shutdown() {
      imageReceiver.stopPlaying();
    }



    @Override
    public Dimension getPreferredSize() {

      if (fullscreen == null) return new Dimension(640, 360);
      return new Dimension(fullscreen.getWidth(), fullscreen.getHeight());

    }



    @Override
    public void paintComponent(Graphics g) {
      super.paintComponent(g);
      g.drawImage(fullscreen, 0, 0, this);
    }

  }

}
