package net.javacomm.restserver;

import jakarta.ws.rs.GET;
import jakarta.ws.rs.POST;
import jakarta.ws.rs.Path;
import jakarta.ws.rs.PathParam;
import jakarta.ws.rs.Produces;
import jakarta.ws.rs.QueryParam;
import jakarta.ws.rs.WebApplicationException;
import jakarta.ws.rs.core.MediaType;
import jakarta.ws.rs.core.Response;
import jakarta.ws.rs.core.Response.Status;
import java.io.InputStream;
import java.sql.SQLException;
import java.util.ArrayList;
import org.apache.logging.log4j.LogManager;
import org.apache.logging.log4j.Logger;
import org.glassfish.jersey.media.multipart.FormDataMultiPart;
import net.javacomm.database.DatabaseException;
import net.javacomm.database.WebdatabaseImpl;
import net.javacomm.share.Constants;



@Path("/chat")
public class ChatFileDownload {

  final static Logger log = LogManager.getLogger(ChatFileDownload.class);
  private WebdatabaseImpl database;

  public ChatFileDownload() {
    database = (WebdatabaseImpl) WebdatabaseImpl.getInstance();
  }



  /**
   * Der Anwender lädt eine Anlage aus einem Besprechungsraum herunter.
   * 
   * 
   * @param anlage
   *               eine Referenz
   * 
   * @return ein Dateistream
   * @throws SQLException
   */
  @GET
  @Path("/anlage")
  @Produces(MediaType.APPLICATION_OCTET_STREAM)
  public InputStream getAnlage(@QueryParam("number") Long anlage, @QueryParam("chunk") Integer chunk) {
    InputStream stream;
    try {
      stream = database.fetchAnlageAsStream(anlage, chunk);
    }
    catch (DatabaseException e) {
      log.info(e.getMessage());
      throw new WebApplicationException();
    }
    return stream;
  }



  /**
   * Der Anwender überträgt alle Attribue der Anlage/Datei. Der Rückgabewert ist
   * ein Schlüssel/Identifier auf die Anlage.
   * 
   * @param multiPart
   *                  alle Attribute
   * 
   * @return Identifier auf die Anlage/Datei
   * @throws SQLException
   */
  @POST
  @Path("save/chatfile/attribute")
  @Produces(MediaType.TEXT_PLAIN)
  public Long postAttributes(FormDataMultiPart multiPart) {
    Long id = null;
    Long filesize = Long.valueOf(multiPart.getField("FILESIZE").getValue());
    String chatid = multiPart.getField("CHATID").getValue();
    String absolute = multiPart.getField("ABSOLUTE").getValue();
    String filename = multiPart.getField("FILENAME").getValue();
    if (filesize > Constants.MAX_FILESIZE) {
      Response response = Response.status(Status.INTERNAL_SERVER_ERROR)
          .entity(
              String.valueOf(filesize) + " > " + String.valueOf(Constants.MAX_FILESIZE) + " - Filesize to big"
          ).type(MediaType.TEXT_PLAIN).build();

      throw new WebApplicationException(response);
    }
    try {
      id = database.saveChatfile(Long.valueOf(filesize), chatid, absolute, filename);
    }
    catch (DatabaseException e) {
      log.warn(e.getMessage());
      throw new WebApplicationException();
    }
    return id;
  }



  /**
   * Der Anwender lädt eine Datei aus einem Privatechat herunter.
   * 
   * @param number
   *               eine Referenz
   * 
   * @return ein Dateistream
   */
  @GET
  @Path("/load")
  @Produces(MediaType.APPLICATION_OCTET_STREAM)
  public InputStream getFileDownload(@QueryParam("number") int number) {
    InputStream stream;
    try {
      stream = database.fetchChatfileAsStream(number);
    }
    catch (SQLException e) {
      log.error(e.getMessage(), e);
      throw new WebApplicationException();
    }
    return stream;
  }



  /**
   * Ermittle die Dateigröße der Anlage.
   * 
   * @param anlage
   *               dieser Datenbankschlüssel verweist auf eine Anlage
   * @return die Dateigröße einer Anlage
   */
  @GET
  @Path("/anlage/size")
  @Produces(MediaType.TEXT_PLAIN)
  public long getFilesize(@QueryParam("anlage") long anlage) {
    long filesize;
    try {
      filesize = database.fetchAnlageSize(anlage);
    }
    catch (DatabaseException e) {
      log.info(e.getMessage());
      throw new WebApplicationException();
    }
    return filesize;

  }



  /**
   * Hole alle Datenblöcke einer Datei.
   * 
   * @param anlage
   *               Datei/Attachment
   * @return {@code ArrayList<Long>}, alle Chunknummern in einem Json Array
   */
  @GET
  @Path("/anlage/chunknumbers/{anlage}")
  @Produces(MediaType.TEXT_PLAIN)
  public String fetchChunknumbers(@PathParam(value = "anlage") long anlage) {
    ArrayList<Long> numbers;
    try {
      numbers = database.fetchChunknumbers(anlage);
    }
    catch (DatabaseException e) {
      log.error(e.getMessage());
      throw new WebApplicationException();
    }
    return Wrapper.toString(numbers);
  }

}
