/*
 * The MIT License
 * Copyright © 2023 Thorsten Schulz <https://javacomm.net>
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy
 * of this software and associated documentation files (the "Software"), to deal
 * in the Software without restriction, including without limitation the rights
 * to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
 * copies of the Software, and to permit persons to whom the Software is
 * furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in
 * all copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
 * THE SOFTWARE.
 */
package net.javacomm.mastermind.gui;

import java.awt.Color;
import java.awt.Component;
import java.awt.FlowLayout;
import java.awt.GridBagConstraints;
import java.awt.GridBagLayout;
import java.beans.PropertyChangeListener;
import java.beans.PropertyChangeSupport;
import java.io.IOException;
import java.util.HashMap;
import java.util.List;
import java.util.Objects;
import javax.sound.sampled.LineUnavailableException;
import javax.sound.sampled.UnsupportedAudioFileException;
import javax.swing.ImageIcon;
import javax.swing.JLabel;
import javax.swing.JPanel;
import javax.swing.JScrollBar;
import javax.swing.JScrollPane;
import javax.swing.ScrollPaneConstants;
import javax.swing.UIManager;
import org.apache.logging.log4j.LogManager;
import org.apache.logging.log4j.Logger;
import org.nexuswob.util.Util;
import net.javacomm.mastermind.audio.SoundClip;
import net.javacomm.mastermind.control.Status;
import net.javacomm.mastermind.control.StatusListener;
import net.javacomm.mastermind.tool.DefColors;
import net.javacomm.mastermind.tool.Knopf;
import net.javacomm.multilingual.Babelfish;
import net.javacomm.multilingual.MultilingualLabel;
import net.javacomm.multilingual.MultilingualString;
import net.javacomm.multilingual.schema.ISO639;
import net.javacomm.multilingual.schema.KEY;



/**
 * Alle Grundelemente für ein Mastermind-Spielfeld.
 * 
 */
@SuppressWarnings("serial")
public abstract class ASpielfeld extends JScrollPane implements Babelfish, StatusListener {
  private final static Logger log = LogManager.getLogger(ASpielfeld.class);

  protected final HashMap<DefColors, DefColors> farbkreis = new HashMap<>(); // gilt für alle
  protected MultilingualString benutzerfehler = new MultilingualString(KEY.STRING_BENUTZERFEHLER);
  protected MultilingualString spielstart = new MultilingualString(KEY.STRING_ES_WURDE_NICHT);
  protected ImageIcon logo = new ImageIcon(getClass().getResource(Constants.LOGO));
  protected ImageIcon sleepman = new ImageIcon(getClass().getResource(Constants.SLEEPMAN));
  protected MultilingualLabel labelRundeZuEnde = new MultilingualLabel(KEY.STRING_DIE_RUNDE_IST);
  protected MultilingualString runde = new MultilingualString(KEY.STRING_RUNDE);
  protected MultilingualLabel verloren = new MultilingualLabel(KEY.LABEL_DU_HAST_VERLOREN);
  protected ImageIcon sadman = new ImageIcon(getClass().getResource(Constants.SADMAN));
  protected MultilingualLabel gewonnen = new MultilingualLabel(KEY.LABEL_DU_HAST_GEWONNEN);
  protected ImageIcon hurraman = new ImageIcon(getClass().getResource(Constants.HURRAMAN));
  protected DuBistInRundePanel duBistInPanel = new DuBistInRundePanel();

  protected GridBagLayout gridbag = new GridBagLayout();
  protected GridBagConstraints con = new GridBagConstraints();
  protected JPanel mainpane = new JPanel();
  protected JScrollBar vertical;
  protected JScrollBar horizontal;

  protected JLabel corner;

  String solutionPattern;

  /**
   * Nicht überschreiben in abgeleiteten Klassen. Falls doch, dann müssen auch die
   * Listener überschrieben werden.
   */
  PropertyChangeSupport changes = new PropertyChangeSupport(this);

  public ASpielfeld() {

    farbkreis.put(DefColors.YELLOW, DefColors.ORANGE);
    farbkreis.put(DefColors.ORANGE, DefColors.RED);
    farbkreis.put(DefColors.RED, DefColors.PURPUR);
    farbkreis.put(DefColors.PURPUR, DefColors.BLUE);
    farbkreis.put(DefColors.BLUE, DefColors.GREEN);
    farbkreis.put(DefColors.GREEN, DefColors.YELLOW);

    corner = new JLabel();
    corner(UIManager.getColor("nimbusSelectionBackground"));
    corner.setOpaque(true);
    setCorner(ScrollPaneConstants.LOWER_RIGHT_CORNER, corner); // ok!

    mainpane.setLayout(gridbag);
    mainpane.setBackground(Constants.OLIV);// damit das 8er-feld keine Streifen erzeugt
    setViewportView(mainpane);
    vertical = getVerticalScrollBar();
    vertical.setUnitIncrement(24);
    horizontal = getHorizontalScrollBar();
    horizontal.setUnitIncrement(24);

  }



  /**
   * Der Name für diese Komponente.
   * 
   * @return ein Bezeichner
   */
  public abstract String getDescription();



  public void addSpielfeldListener(PropertyChangeListener listener) {
    changes.addPropertyChangeListener(listener);
  }



  public void removeSpielfeldListener(PropertyChangeListener listener) {
    changes.removePropertyChangeListener(listener);
  }



  public void removeAllListener() {
    for (PropertyChangeListener pl : changes.getPropertyChangeListeners()) {
      changes.removePropertyChangeListener(pl);
    }
  }



  /**
   * Eine Hilfsmehode, welche die rechte untere Ecke in der Videoansicht einfärbt.
   * Nach einem L&F-Wechsel wurde die Farbe nicht geändert.
   * 
   * @param color
   *              eine Farbe
   */
  public void corner(Color color) {
    corner.setBackground(Constants.OLIV);
  }



  /**
   * Das Spielfeld wird auf den Status
   * {@link net.javacomm.mastermind.control.Status#Init} zurückgesetzt. Alle
   * Anzeigen werden gelöscht und der Anwender kann Spieleinstellungen vornehmen.
   */
  public abstract void reset();

  /**
   * Die höchstmögliche Rundenzahl in einem Spiel.
   * 
   * @return höchstmögliche Rundenzahl
   * 
   */
  public abstract int getRounds();

  /**
   * Einem Spielfeld wird immer ein <code>Status</code> zugeordnet, in dem es sich
   * befindet.
   * 
   * @return in diesem {@link net.javacomm.mastermind.control.Status} befindet
   *         sich das Spielfeld
   */
  public abstract Status hasStatus();



  /**
   * Das Lösungsmuster wird übergeben.
   * 
   * @param pattern
   *                Lösungsmuster
   */
  public void setSolutionPattern(String pattern) {
    solutionPattern = pattern;
  }



  /**
   * Das Lösungsmuster wird zurückgegeben.
   * 
   * @return das Lösungsmuster ist ein Akronym aus 4 Buchstaben
   */
  public String getSolutionPattern() {
    return solutionPattern;
  }



  /**
   * Der Anwender hat auf der Rundenknopfleiste diese Farbe ausgewählt. Dadurch
   * wird die Farbe zur aktuellen Farbe auf der Farbvorauswahlleiste.
   * 
   * @return diese Farbe
   */
  public abstract DefColors getAktuelleFarbe();

  /**
   * Der Anwender hat auf der Farbvorauswahlleiste eine Farbe markiert.
   * 
   * @param color
   *              diese Farbe
   */
  public abstract void setAktuelleFarbe(DefColors color);



  /**
   * Audio-Click für einen Eingabefehler.
   */
  public void errorClip() {
    try {
      SoundClip.playClickError();
    }
    catch (IOException | UnsupportedAudioFileException | LineUnavailableException e) {
      log.info(e.getMessage(), e);
    }
  }



  /**
   * Audio-Click für einen Eingabefehler.
   */
  public void sleepClip() {
    try {
      SoundClip.playMaleTiredYawn();
    }
    catch (IOException | UnsupportedAudioFileException | LineUnavailableException e) {
      log.info(e.getMessage(), e);
    }
  }



  /**
   * Scrolle in die 1-te Reihe.
   */
  public void scrollToFirstRow() {
    validate();
    revalidate();
    vertical.setValue(vertical.getMaximum());
  }



  /**
   * Spielt Verloren-Audio.
   */
  public void gameOver() {
    try {
      SoundClip.playCartoonGossip();
    }
    catch (IOException | UnsupportedAudioFileException | LineUnavailableException e) {}
  }



  /**
   * Spielt Gewonnen-Audio.
   */
  public void victory() {
    try {
      SoundClip.playVictory();
    }
    catch (IOException | UnsupportedAudioFileException | LineUnavailableException e) {}
  }



  @Override
  public void setLanguage(ISO639 code) {
    benutzerfehler.setLanguage(code);
    spielstart.setLanguage(code);
    duBistInPanel.setLanguage(code);
    labelRundeZuEnde.setLanguage(code);
    runde.setLanguage(code);
    verloren.setLanguage(code);
    gewonnen.setLanguage(code);
  }

  /**
   * Eine Fehlermeldung für einen Dialog.
   */
  @SuppressWarnings({"unused"})
  protected static class DuBistInRundePanel extends JPanel implements Babelfish {

    /**
     * 
     */
    private static final long serialVersionUID = -1677130394041955273L;
    private FlowLayout flow = new FlowLayout(FlowLayout.LEFT);
    private MultilingualLabel labelRunde = new MultilingualLabel(KEY.LABEL_DU_BIST_IN_RUNDE);

    DuBistInRundePanel() {
      add(labelRunde);
    }



    public void setKnopfAufRundenzahl(int nummer) {
      List<Component> liste = Util.getAllComponents(this);
      liste.forEach(component -> {
        if (component instanceof JNumberButton button) {
          remove(button);
        }
      });
      JNumberButton numberButton = new JNumberButton(nummer);
      numberButton.removeAllListener();
      add(numberButton);
    }



    @Override
    public void setLanguage(ISO639 code) {
      labelRunde.setLanguage(code);
    }

  }

  protected static class KnopfUndFarbe {
    private DefColors farbe;
    private Knopf knopf;

    DefColors getFarbe() {
      return farbe;
    }



    Knopf getKnopf() {
      return knopf;
    }



    void setFarbe(DefColors farbe) {
      this.farbe = farbe;
    }



    void setKnopf(Knopf knopf) {
      this.knopf = knopf;
    }



    @Override
    public int hashCode() {
      return Objects.hash(farbe, knopf);
    }



    @Override
    public boolean equals(Object obj) {
      if (this == obj) return true;
      if (obj == null) return false;
      if (getClass() != obj.getClass()) return false;
      KnopfUndFarbe other = (KnopfUndFarbe) obj;
      return farbe == other.farbe && knopf == other.knopf;
    }
  }

}
