/*
 * The MIT License
 * Copyright © 2023 Thorsten Schulz <https://javacomm.net>
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy
 * of this software and associated documentation files (the "Software"), to deal
 * in the Software without restriction, including without limitation the rights
 * to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
 * copies of the Software, and to permit persons to whom the Software is
 * furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in
 * all copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
 * THE SOFTWARE.
 */
package net.javacomm.mastermind.gui;

import java.awt.Color;
import java.awt.Dimension;
import java.beans.PropertyChangeListener;
import java.beans.PropertyChangeSupport;
import java.beans.PropertyVetoException;
import java.net.URL;
import javax.swing.ImageIcon;
import javax.swing.JSplitPane;
import javax.swing.WindowConstants;
import javax.swing.event.InternalFrameAdapter;
import javax.swing.event.InternalFrameEvent;
import org.apache.logging.log4j.LogManager;
import org.apache.logging.log4j.Logger;
import net.javacomm.multilingual.Babelfish;
import net.javacomm.multilingual.schema.ISO639;
import net.javacomm.multilingual.schema.KEY;
import net.javacomm.window.manager.Control;
import net.javacomm.window.manager.Frames;
import net.javacomm.window.manager.WM;
import net.javacomm.window.manager.WindowAnrufenEvent;
import net.javacomm.window.manager.WindowAnrufenListener;
import net.javacomm.window.manager.WindowManagerInternalFrame;



/******************************************************************/
/*                                                                */
/*                      Class Mastermind                          */
/*                                                                */
/******************************************************************/

public class JMastermindFrame extends WindowManagerInternalFrame implements Babelfish, WindowAnrufenListener {

  private final static Logger log = LogManager.getLogger(JMastermindFrame.class);

  private static final long serialVersionUID = -4206330791001343460L;
  public final static int SCREEN_WIDTH = 1118; // 640 + 271
  public final static int SCREEN_HEIGHT = 650; // 360+24
  private InternalFrameAction internalAction = new InternalFrameAction();
  private ImageIcon icon;
  private URL url;
  private JGamePanel gamepanel = new JGamePanel();
  private PropertyChangeSupport changes = new PropertyChangeSupport(this);
  protected String extText;
  private JSplitPane splitpane = new JSplitPane(JSplitPane.HORIZONTAL_SPLIT);

  public JMastermindFrame() {
    setContentPane(splitpane);

    url = getClass().getResource(Constants.LOGO);
    icon = new ImageIcon(url);
    setFrameIcon(icon);
    setMaximizable(true);
    setResizable(true);
    setIconifiable(false);
    setClosable(true);
    setDefaultCloseOperation(WindowConstants.DISPOSE_ON_CLOSE);
    addInternalFrameListener(internalAction);
    wm.setMinimumSize(new Dimension(0, 0));
    wm.addWMListener((propertyEvent) -> {
      changes.firePropertyChange(propertyEvent);
    });

    gamepanel.setMinimumSize(new Dimension(0, 0));

    splitpane.setRightComponent(wm);
    splitpane.setLeftComponent(gamepanel);
    splitpane.setDividerLocation(WIDTH - WindowManagerInternalFrame.SPLITWIDTH);
    splitpane.setOneTouchExpandable(true);

    // Frame Parameter

    setSize(SCREEN_WIDTH, SCREEN_HEIGHT);
  }



  /**
   * Die Trennleiste wird in horizontaler Richtung positioniert. Die Angabe
   * erfolgt in Prozent. 0-Prozent ist ganz links und 100-Prozent ist ganz rechts.
   *
   * @param prozent
   *                liegt im Intervall [0..1]
   */
  public void setHorizontalDivider(double prozent) {
    if (0 > prozent || prozent > 1) {
      log.warn(prozent + " - die Prozentangabe liegt nicht im Intervall [0..1]");
    }
    splitpane.setDividerLocation(prozent);
  }



  public void addMastermindListener(PropertyChangeListener listener) {
    changes.addPropertyChangeListener(listener);
  }



  public void removeMastermindListener(PropertyChangeListener listener) {
    changes.removePropertyChangeListener(listener);
  }



  public void removeAllListener() {
    gamepanel.removeAllListener();
    for (PropertyChangeListener pl : changes.getPropertyChangeListeners()) {
      removeMastermindListener(pl);
    }
    wm.removeAllListener();
    removeInternalFrameListener(internalAction);
  }



  /**
   * Mastermindfenster wird geschlossen.
   */
  public void closing() {
    try {
      setClosed(true);
    }
    catch (PropertyVetoException e) {}
  }



  @Override
  public Dimension getMinimumSize() {
    return new Dimension(SCREEN_WIDTH - 100, 650);
  }



  @Override
  public void activated() {
    try {
      setSelected(true);
      wm.setActivNode(getFrameTitleId());
    }
    catch (PropertyVetoException e) {}
  }



  @Override
  public String getFrameTitleId() {
    return getTitle();
  }



  @Override
  public void setFrameTitleId(String str) {
    setTitle(str);
  }



  @Override
  public Frames getType() {
    return Frames.MASTERMIND;
  }



  @Override
  public void insertWM(WM record) {
    wm.addWindow(record);
  }



  @Override
  public void removeWM(WM record) {
    wm.removeWM(record);
  }



  @Override
  public void setLanguage(ISO639 code) {
    gamepanel.setLanguage(code);
    setLanguage(KEY.FRAME_MASTERMIND, code); // der neue Name wird gesetzt
    treeView(code);
  }

  /* Innerclass WindowAction */

  class InternalFrameAction extends InternalFrameAdapter {

    @Override
    public void internalFrameClosed(InternalFrameEvent event) {
      changes.firePropertyChange(JMastermindFrame.class.getName(), Control.NULL, Control.CLOSE);
    }

  }

  /**
   * Das Gamepanel soll farblich einheitlich erscheinen. Die Farbe kommt von der
   * eingestellten Eissorte für die Oberfläche.
   * 
   * @param color
   *              Eckfarbe
   */
  public void corner(Color color) {
    gamepanel.corner(color);
  }



  @Override
  public void ringRing(WindowAnrufenEvent event) {
    extText = event.getNickname();
    getMultiAbnehmen().setLanguage(KEY.FRAME_ABNEHMEN, event.getCode(), extText);
    getMultiAnrufen().setLanguage(KEY.TELEFON_ANRUFEN, event.getCode(), extText);
  }



  /**
   * In der Grundeinstellung für jedes Spiel soll der vertikale Scrollbalken ganz
   * unten sein, so dass die 1-te Spielrunde angezeigt wird.
   * 
   */
  public void scrollToFirstRow() {
    gamepanel.scrollToFirstRow();
  }

}