/*
 * The MIT License
 * Copyright © 2023 Thorsten Schulz <https://javacomm.net>
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy
 * of this software and associated documentation files (the "Software"), to deal
 * in the Software without restriction, including without limitation the rights
 * to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
 * copies of the Software, and to permit persons to whom the Software is
 * furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in
 * all copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
 * THE SOFTWARE.
 */
package net.javacomm.mastermind.tool;

import java.util.ArrayList;
import net.javacomm.mastermind.control.MastermindException;
import net.javacomm.mastermind.gui.Constants;



public final class PatternChecker {

  private PatternChecker() {}



  /**
   * Das Versuchsmuster wird gegen das Lösungsmuster geprüft.
   * 
   * 
   * @param attemptPattern
   *                        Versuchsmuster
   * @param solutionPattern
   *                        Lösungsmuster
   * @return Trefferbericht
   * @throws MastermindException
   */
  public static String roundUpReport(String attemptPattern, String solutionPattern)
      throws MastermindException {
    if (
      attemptPattern.length() != Knopf.values().length
    ) throw new MastermindException(
        String.valueOf(attemptPattern.length()) + " - das Versuchsmuster hat eine unzulässige Länge"
    );
    if (
      solutionPattern.length() != Knopf.values().length
    ) throw new MastermindException(
        String.valueOf(solutionPattern.length()) + " - das Lösungsmuster hat eine unzulässige Länge"
    );
    int countWhite = 0;
    int countBrown = 0;
    int countBlack = 0; // countBlack = 4- countWhite-countBrown

    // input sammelt geprüfte Farben
    ArrayList<Character> input = new ArrayList<>();
    int count = attemptPattern.length();
    // weiße Stifte
    for (int position = 0; position < count; position++) {
      char knopf = attemptPattern.charAt(position);
      if (knopf == solutionPattern.charAt(position)) {
        input.add(knopf); // jeder Knopf aus dem attempPattern, wird in input eingetragen
        countWhite++;
      }
    }

    // braune Stifte
    int solutionLength = solutionPattern.length();
    for (int position = 0; position < count; position++) { // durchlaufe das AttemptPattern von 0..n
      char knopfAusAttempt = attemptPattern.charAt(position); // knopf_1, knopf_2, ...
      if (input.contains(knopfAusAttempt)) continue; // die Farbe kam bereits in einem Knopf vor hole nächsten
                                                     // Knopf
      // könnte ein brauner Knopf werden, wenn die Farbe in der Benutzereingabe(input)
      // noch nicht vorkam
      input.add(knopfAusAttempt); // jeder Knopf aus dem attempPattern, wird in input eingetragen
      for (int solutionPosition = 0; solutionPosition < solutionLength; solutionPosition++) {
        if (knopfAusAttempt == solutionPattern.charAt(solutionPosition)) {
          // hurra ich bin braun
          countBrown++;
          // aus der for-schleife springen und den nächsten Knopf prüfen
          break;
        }
      }
    }
    countBlack = Knopf.values().length - countWhite - countBrown;
    if (countBlack < 0)
      throw new UnsupportedOperationException(String.valueOf(countBlack) + " - das ist ein böser Fehler");

    StringBuffer rundenbericht = new StringBuffer();
    for (int white = 0; white < countWhite; white++) {
      rundenbericht.append(DefColors.WHITE.charValue());
    }
    for (int brown = 0; brown < countBrown; brown++) {
      rundenbericht.append(DefColors.BROWN.charValue());
    }
    for (int black = 0; black < countBlack; black++) {
      rundenbericht.append(DefColors.BLACK.charValue());
    }
    return rundenbericht.toString();
  }



  /**
   * Das Versuchsmuster wird gegen das Lösungsmuster geprüft.
   * 
   * 
   * @param test
   *               Versuchmsuster
   * @param result
   *               Lösungsmuster
   * @return Trefferbericht
   * 
   * @throws EmptyResultException
   * @throws PatternToShortException
   */
  public static String checkPattern(String test, String result)
      throws EmptyResultException, PatternToShortException {
    String report = "";

    if (result.length() == 0) {
      throw new EmptyResultException();
    }

    if (test.length() < Constants.COUNT_PATTERN) {
      throw new PatternToShortException();
    }

    char[] testLetter = test.toCharArray();
    char[] resultLetter = result.toCharArray();

    /* Weiße Stifte */
    for (int i = 0; i < testLetter.length; i++) {
      if (testLetter[i] == resultLetter[i]) {
        testLetter[i] = 'ö'; // Muster aus dem Spiel nehmen, da es ja schon berücksichtigt worden ist
        resultLetter[i] = 'ä'; // Muster aus dem Spiel nehmen, da es ja schon berücksichtigt worden ist
//        report += DefColors.CHAR_WHITE;
        report += DefColors.WHITE.charValue();
      }
    }

    /* Braune Stifte */
    for (int i = 0; i < resultLetter.length; i++) {
      for (int j = 0; j < testLetter.length; j++) {
        if (testLetter[j] == resultLetter[i]) {
          testLetter[j] = 'ö'; // Muster aus dem Spiel nehmen, da es ja schon berücksichtigt worden ist
          resultLetter[i] = 'ä'; // Muster aus dem Spiel nehmen, da es ja schon berücksichtigt worden ist
//          report += DefColors.CHAR_BROWN;
          report += DefColors.BROWN.charValue();
        }
      }
    }

    while (report.length() < 4) {
//      report += DefColors.CHAR_BLACK;
      report += DefColors.BLACK.charValue();
    }

    return report;
  }

}
